import {
  faMagnifyingGlassMinus,
  faMagnifyingGlassPlus,
  faSlidersSimple,
  faVolume,
  faVolumeSlash,
  faXmark,
} from "@fortawesome/pro-regular-svg-icons";
import {
  Match,
  Setter,
  Show,
  Switch,
  createEffect,
  createSignal,
  onCleanup,
} from "solid-js";
import { FontAwesomeIcon } from "./FontAwesomeIcon";

interface Asset {
  src: string;
}

function AssetHandler() {
  const [moving, setMoving] = createSignal(false);
  const [active, setActive] = createSignal("");
  const [zoomLevel, setZoomLevel] = createSignal(0);
  const zoomAmount = () => 1 + zoomLevel() * 0.5;
  const [x, setX] = createSignal(0);
  const [y, setY] = createSignal(0);

  function FullscreenView() {
    const [muted, setMuted] = createSignal(true);
    const [controlls, setControlls] = createSignal(false);
    const [clientX, setClientX] = createSignal(0);
    const [clientY, setClientY] = createSignal(0);
    const [ref, setRef] = createSignal<HTMLElement | null>(null);

    const handleMouseMove = (event: MouseEvent) => {
      const { clientX, clientY } = event;
      setClientX(clientX);
      setClientY(clientY);
    };

    createEffect(() => {
      const thisRef = ref();
      if (!thisRef) return;
      const rectA = thisRef.getBoundingClientRect();
      const origSizeW = rectA.width / zoomAmount();
      const origSizeH = rectA.height / zoomAmount();
      const rectW = wrapperRef.getBoundingClientRect();
      const hori = rectW.width - origSizeW + origSizeW / zoomAmount();
      const verti = rectW.height - origSizeH + origSizeH / zoomAmount();
      const xP = (clientX() - hori / 2) / (rectW.width - hori);
      const yP = (clientY() - verti / 2) / (rectW.height - verti);
      setX((xP > 0 ? (xP < 1 ? xP : 1) : 0) * origSizeW);
      setY((yP > 0 ? (yP < 1 ? yP : 1) : 0) * origSizeH);
      const moving =
        clientX() > (rectW.width - origSizeW) / 2 &&
        clientX() < rectW.width - (rectW.width - origSizeW) / 2 &&
        clientY() > (rectW.height - origSizeH) / 2 &&
        clientY() < rectW.height - (rectW.height - origSizeH) / 2;
      setMoving(moving);
    });

    let wrapperRef: HTMLDivElement;

    createEffect(() => {
      if (active()) return;
      setZoomLevel(0);
      setMuted(true);
      setControlls(false);
    });

    const handleKeyDown = (event: KeyboardEvent) => {
      if (!active() || "Escape" != event.key) return;
      event.preventDefault();
      setActive("");
    };

    const handleWheel = (event: WheelEvent) => {
      if (!active()) return;
      event.preventDefault();
      if (event.deltaY > 0 && !(zoomLevel() <= 0)) setZoomLevel((e) => e - 1);
      else if (event.deltaY < 0 && !(zoomLevel() >= 4))
        setZoomLevel((e) => e + 1);
    };

    createEffect(() => {
      window.addEventListener("keydown", handleKeyDown);
      window.addEventListener("wheel", handleWheel, { passive: false });
      onCleanup(() => {
        window.removeEventListener("keydown", handleKeyDown);
        window.removeEventListener("wheel", handleWheel);
      });
    });

    return (
      <div
        ref={wrapperRef!}
        onClick={(e) => e.target === wrapperRef && setActive("")}
        onMouseMove={handleMouseMove}
        classList={{ "fullscreen-asset": true, active: !!active() }}
      >
        <Asset
          src={active()}
          volume={!muted()}
          controlls={controlls()}
          setRef={setRef}
        />
        <div class="controlls">
          <button onClick={() => setActive("")} title="Close">
            <FontAwesomeIcon icon={faXmark} />
          </button>
          <Show when={active().startsWith("/videos")}>
            <button
              onClick={() => setMuted((e) => !e)}
              style={{ "border-style": controlls() ? "dashed" : "" }}
              title={muted() ? "Unmute" : "Mute"}
            >
              <FontAwesomeIcon icon={!muted() ? faVolume : faVolumeSlash} />
            </button>
            <button
              onClick={() => setControlls((e) => !e)}
              style={{ "border-style": controlls() ? "inset" : "" }}
              title="Show controlls"
            >
              <FontAwesomeIcon icon={faSlidersSimple} />
            </button>
          </Show>
          <button
            onClick={() => setZoomLevel((e) => e + 1)}
            disabled={zoomLevel() >= 4}
            title="Zoom +"
          >
            <FontAwesomeIcon icon={faMagnifyingGlassPlus} />
          </button>
          <button
            onClick={() => setZoomLevel((e) => e - 1)}
            disabled={zoomLevel() <= 0}
            title="Zoom -"
          >
            <FontAwesomeIcon icon={faMagnifyingGlassMinus} />
          </button>
        </div>
      </div>
    );
  }

  function Asset(
    props: Asset & {
      class?: string;
      volume?: boolean;
      controlls?: boolean;
      setRef?: Setter<HTMLElement | null>;
    },
  ) {
    let imgRef: HTMLImageElement;
    let videoRef: HTMLVideoElement;

    const shouldZoom = () => active() === props.src && !!zoomLevel();

    return (
      <Switch>
        <Match when={props.src.startsWith("/images")}>
          <img
            ref={imgRef!}
            style={
              shouldZoom()
                ? {
                    "--zoom": zoomAmount(),
                    "--x": x() + "px",
                    "--y": y() + "px",
                  }
                : {}
            }
            onMouseEnter={() => props.setRef && props.setRef(imgRef)}
            classList={{
              [props.class ?? ""]: !!props.class,
              zoomed: shouldZoom(),
              move: moving(),
            }}
            onClick={() => setActive(props.src)}
            src={props.src}
            alt={props.src}
          />
        </Match>
        <Match when={props.src.startsWith("/videos")}>
          <video
            ref={videoRef!}
            onMouseEnter={() => props.setRef && props.setRef(videoRef)}
            style={
              shouldZoom()
                ? {
                    "--zoom": zoomAmount(),
                    "--x": x() + "px",
                    "--y": y() + "px",
                  }
                : {}
            }
            classList={{
              [props.class ?? ""]: !!props.class,
              zoomed: shouldZoom(),
              move: moving(),
              active: !props.setRef && active() === props.src,
            }}
            onClick={() => setActive(props.src)}
            width="auto"
            height="auto"
            autoplay
            muted={!props.volume}
            loop
            controls={props.controlls}
          >
            <source src={props.src} type="video/mp4" />
          </video>
        </Match>
      </Switch>
    );
  }

  return { FullscreenView, Asset };
}

export default AssetHandler;
